from typing import List, Optional

from bullhorn.types import client_contact, client_corporation, corporate_user, lead


class Opportunity:
    id: int
    actualCloseDate: Optional[int]
    address: str
    appointments: List[int]
    assignedDate: Optional[int]
    assignedUsers: List[int]
    benefits: str
    billRateCategoryID: int
    bonusPackage: str
    branchCode: str
    businessSector: int
    businessSectors: List[int]
    campaignSource: str
    category: int
    categories: List[int]
    certifications: List[int]
    clientContact: client_contact.ClientContact
    clientCorporation: client_corporation.ClientCorporation
    committed: bool
    customDate1: Optional[float]
    customDate2: Optional[float]
    customDate3: Optional[float]
    customFloat1: Optional[float]
    customFloat2: Optional[float]
    customFloat3: Optional[float]
    customInt1: Optional[int]
    customInt2: Optional[int]
    customInt3: Optional[int]
    customText1: Optional[str]
    customText2: Optional[str]
    customText3: Optional[str]
    customText4: Optional[str]
    customText5: Optional[str]
    customText6: Optional[str]
    customText7: Optional[str]
    customText8: Optional[str]
    customText9: Optional[str]
    customText10: Optional[str]
    customTextBlock1: Optional[str]
    customTextBlock2: Optional[str]
    customTextBlock3: Optional[str]
    customTextBlock4: Optional[str]
    customTextBlock5: Optional[str]
    customObject1s: List[int]
    customObject2s: List[int]
    customObject3s: List[int]
    customObject4s: List[int]
    customObject5s: List[int]
    customObject6s: List[int]
    customObject7s: List[int]
    customObject8s: List[int]
    customObject9s: List[int]
    customObject10s: List[int]
    dateAdded: int
    dateLastModified: int
    dealValue: float
    degreeList: str
    description: str
    educationDegree: str
    effectiveDate: Optional[int]
    estimatedDuration: Optional[float]
    estimatedEndDate: Optional[int]
    estimatedHoursPerWeek: Optional[float]
    estimatedStartDate: int
    estimatedBillRate: float
    expectedCloseDate: Optional[int]
    expectedFee: float
    expectedPayRate: float
    externalCategoryID: int
    externalID: str
    history: List[int]
    ignoreUntilDate: Optional[int]
    isDeleted: bool
    isOpen: bool
    jobOrders: List[int]
    lead: Optional[lead.Lead]
    markUpPercentage: Optional[float]
    notes: List[int]
    numOpenings: int
    onSite: str
    optionsPackage: str
    owner: corporate_user.CorporateUser
    priority: int
    publicDescription: str
    publishedZip: str
    reasonClosed: str
    salary: float
    salaryUnit: str
    skillList: str
    skills: List[int]
    source: str
    specialties: List[int]
    status: str
    tasks: List[int]
    taxRate: float
    taxStatus: str
    tearsheets: List[int]
    title: str
    type: str
    weightedDealValue: float
    willRelocate: bool
    winProbabilityPercent: Optional[float]
    yearsRequired: int



import os

from bullhorn.client import BullhornClient


def get_placements_and_candidates():
    # Set credentials
    session_token = os.environ.get("BULLHORN_SESSION_TOKEN")
    rest_url = os.environ.get("BULLHORN_REST_URL")

    # Initialise client
    bc = BullhornClient(
        token=session_token,
        rest_url=rest_url,
    )

    # Get placements result
    placements = bc.get_placements(
        query="dateLastModified:{2023/01/01 TO *}",
        fields="id,candidate,comments,dateAdded,dateBegin,dateEnd,dateLastModified,referralFee,salary,status",
    )
    print("Success: %r" % (placements))

    # Get candidates result
    candidates = bc.get_candidates(
        query="dateLastModified:{2023/01/01 TO *}",
        fields="id,firstName,lastName",
    )
    print("Success: %r" % (candidates))



from decimal import Decimal
from typing import Any, List, Optional

from bullhorn.types import category, corporate_user, person


class Candidate(person.Person):
    id: int
    address: str
    blacklistClientCorporations: List[int]
    businessSectors: List[int]
    candidateSource: str
    category: category.Category
    categories: List[int]
    certifications: str
    clientRating: Optional[int]
    comments: str
    companyName: str
    companyURL: str
    customDate1: Optional[int]
    customDate2: Optional[int]
    customDate3: Optional[int]
    customDate4: Optional[int]
    customDate5: Optional[int]
    customDate6: Optional[int]
    customDate7: Optional[int]
    customDate8: Optional[int]
    customDate9: Optional[int]
    customDate10: Optional[int]
    customDate11: Optional[int]
    customDate12: Optional[int]
    customDate13: Optional[int]
    customEncryptedText1: str
    customEncryptedText2: str
    customEncryptedText3: str
    customEncryptedText4: str
    customEncryptedText5: str
    customEncryptedText6: str
    customEncryptedText7: str
    customEncryptedText8: str
    customEncryptedText9: str
    customEncryptedText10: str
    customFloat1: Optional[float]
    customFloat2: Optional[float]
    customFloat3: Optional[float]
    customFloat4: Optional[float]
    customFloat5: Optional[float]
    customFloat6: Optional[float]
    customFloat7: Optional[float]
    customFloat8: Optional[float]
    customFloat9: Optional[float]
    customFloat10: Optional[float]
    customFloat11: Optional[float]
    customFloat12: Optional[float]
    customFloat13: Optional[float]
    customFloat14: Optional[float]
    customFloat15: Optional[float]
    customFloat16: Optional[float]
    customFloat17: Optional[float]
    customFloat18: Optional[float]
    customFloat19: Optional[float]
    customFloat20: Optional[float]
    customFloat21: Optional[float]
    customFloat22: Optional[float]
    customFloat23: Optional[float]
    customInt1: Optional[int]
    customInt2: Optional[int]
    customInt3: Optional[int]
    customInt4: Optional[int]
    customInt5: Optional[int]
    customInt6: Optional[int]
    customInt7: Optional[int]
    customInt8: Optional[int]
    customInt9: Optional[int]
    customInt10: Optional[int]
    customInt11: Optional[int]
    customInt12: Optional[int]
    customInt13: Optional[int]
    customInt14: Optional[int]
    customInt15: Optional[int]
    customInt16: Optional[int]
    customInt17: Optional[int]
    customInt18: Optional[int]
    customInt19: Optional[int]
    customInt20: Optional[int]
    customInt21: Optional[int]
    customInt22: Optional[int]
    customInt23: Optional[int]
    customObject1s: List[Any]
    customObject2s: List[Any]
    customObject3s: List[Any]
    customObject4s: List[Any]
    customObject5s: List[Any]
    customObject6s: List[Any]
    customObject7s: List[Any]
    customObject8s: List[Any]
    customObject9s: List[Any]
    customObject10s: List[Any]
    customText1: Optional[str]
    customText2: Optional[str]
    customText3: Optional[str]
    customText4: Optional[str]
    customText5: Optional[str]
    customText6: Optional[str]
    customText7: Optional[str]
    customText8: Optional[str]
    customText9: Optional[str]
    customText10: Optional[str]
    customText11: Optional[str]
    customText12: Optional[str]
    customText13: Optional[str]
    customText14: Optional[str]
    customText15: Optional[str]
    customText16: Optional[str]
    customText17: Optional[str]
    customText18: Optional[str]
    customText19: Optional[str]
    customText20: Optional[str]
    customText21: Optional[str]
    customText22: Optional[str]
    customText23: Optional[str]
    customTextBlock1: Optional[str]
    customTextBlock2: Optional[str]
    customTextBlock3: Optional[str]
    customTextBlock4: Optional[str]
    customTextBlock5: Optional[str]
    customTextBlock6: Optional[str]
    customTextBlock7: Optional[str]
    customTextBlock8: Optional[str]
    customTextBlock9: Optional[str]
    customTextBlock10: Optional[str]
    dateAdded: int
    dateAvailable: Optional[int]
    dateAvailableEnd: Optional[int]
    dateI9Expiration: Optional[int]
    dateLastComment: Optional[int]
    dateLastModified: int
    dateNextCall: Optional[int]
    dateOfBirth: Optional[int]
    dayRate: Optional[Decimal]
    dayRateLow: Optional[Decimal]
    degreeList: str
    description: str
    desiredLocations: str
    disability: str
    educationDegree: str
    email: Optional[str]
    email2: Optional[str]
    email3: Optional[str]
    employeeType: str
    employmentPreference: str
    ethnicity: str
    experience: Optional[int]
    externalID: str
    fax: Optional[str]
    fax2: Optional[str]
    fax3: Optional[str]
    federalAdditionalWithholdingsAmount: Optional[Decimal]
    federalExemptions: Optional[int]
    federalExtraWithholdingAmount: Optional[Decimal]
    federalFilingStatus: Optional[str]
    firstName: str
    gender: str
    hourlyRate: Optional[Decimal]
    hourlyRateLow: Optional[Decimal]
    i9OnFile: Optional[int]
    interviews: List[int]
    isAnonymized: bool
    isDeleted: bool
    isEditable: bool
    isExempt: Optional[bool]
    lastName: str
    leads: List[int]
    linkedPerson: Optional[person.Person]
    localAdditionalWithholdingsAmount: Optional[Decimal]
    localExemptions: Optional[int]
    localFilingStatus: Optional[str]
    localTaxCode: Optional[str]
    maritalStatus: str
    massMailOptOut: bool
    middleName: Optional[str]
    mobile: Optional[str]
    name: str
    namePrefix: str
    nameSuffix: str
    nickName: Optional[str]
    numCategories: Optional[int]
    numOwners: Optional[int]
    occupation: str
    onboardingDocumentReceivedCount: Optional[int]
    onboardingDocumentSentCount: Optional[int]
    onboardingPercentComplete: Optional[int]
    onboardingReceivedSent: Any
    onboardingStatus: str
    otherDeductionsAmount: Optional[Decimal]
    otherIncomeAmount: Optional[Decimal]
    owner: corporate_user.CorporateUser
    pager: Optional[str]
    paperWorkOnFile: str
    password: str
    payrollClientStartDate: Optional[int]
    payrollStatus: str
    phone: Optional[str]
    phone2: Optional[str]
    phone3: Optional[str]
    placements: List[int]
    preferredContact: str
    primarySkills: List[int]
    recentClientList: str
    referredBy: Optional[str]
    referredByPerson: Optional[person.Person]
    salary: Optional[Decimal]
    salaryLow: Optional[Decimal]
    secondaryAddress: str
    secondaryOwners: List[int]
    secondarySkills: List[int]
    sendouts: List[int]
    skillSet: str
    smsOptIn: bool
    source: str
    specialties: List[int]
    ssn: str
    stateAdditionalWithholdingsAmount: Optional[Decimal]
    stateExemptions: Optional[int]
    stateFilingStatus: Optional[str]
    status: str
    submissions: List[int]
    tasks: List[int]
    taxID: str
    taxState: str
    timeZoneOffsetEST: int
    tobaccoUser: str
    totalDependentClaimAmount: Optional[Decimal]
    travelLimit: int
    travelMethod: str
    twoJobs: bool
    type: str
    userDateAdded: int
    username: str
    veteran: str
    webResponses: List[int]
    whitelistClientCorporations: List[int]
    willRelocate: bool
    workAuthorized: bool
    workPhone: str



from bullhorn.types import workers_compensation


class WorkersCompensationRate:
    id: int
    compensation: workers_compensation.WorkersCompensation
    endDate: int
    rate: float
    startDate: int



import json
import logging
import sys
import time
from typing import Any, Dict, List, Optional, Union

import requests

from bullhorn import __version__
from bullhorn.exceptions import BullhornServerError, Forbidden, HTTPException, NotFound
from bullhorn.route import Route
from bullhorn.types import (
    candidate,
    client_contact,
    client_corporation,
    corporate_user,
    job_order,
    job_submission,
    ping,
    placement,
    placement_commission,
)

logger = logging.getLogger(__name__)


class BullhornClient:
    def __init__(
        self,
        token: Optional[str] = None,
        rest_url: Optional[str] = None,
    ) -> None:
        # Set token and namespace
        self.token: Optional[str] = token
        self.rest_url: Optional[str] = rest_url
        # Set user agent
        self.user_agent: str = self._get_user_agent()

    def _get_user_agent(
        self,
    ) -> str:
        """
        Get the user agent string for the HTTP client.

        Returns:
            str:
                The user agent string combining information about the Python wrapper,
                its version, the Python interpreter version, and the aiohttp version.
        """
        version_info = sys.version_info
        user_agent_strings = [
            "Python wrapper for Bullhorn API",
            f"(https://github.com/recruithub/bullhorn {__version__})",
            f"Python/{version_info.major}.{version_info.minor}.{version_info.micro}",
            f"requests/{requests.__version__}",
        ]
        user_agent = " ".join(user_agent_strings)
        return user_agent

    def request(
        self,
        route: Route,
        json: Optional[Dict] = None,
        files=None,
        form=None,
    ) -> Any:
        # Get attributes from route
        method = route.method
        url = route.url
        # Set headers
        headers: Dict[str, str] = {}
        headers["User-Agent"] = self.user_agent
        if self.token is not None:
            headers["BhRestToken"] = self.token
        if json is not None:
            headers["Content-Type"] = "application/json"

        # Fetch response
        data: Optional[Union[Dict[str, Any], str]] = None
        for tries in range(5):
            # Execute request
            try:
                response = requests.request(
                    method=method, url=url, headers=headers, json=json
                )
                logger.debug(
                    f"{method} {url} with {data} has returned {response.status_code}"
                )
                data = response.json()
                # Successful request
                if 300 > response.status_code >= 200:
                    logger.debug(f"{method} {url} has received {data}")
                    return data
                # Server error, so retry request with exponential back-off
                if response.status_code in {500, 502, 503, 504, 524}:
                    time.sleep(1 + tries * 2)
                    continue
                # Client error, so raise exception
                if response.status_code == 403:
                    raise Forbidden(response, data)
                elif response.status_code == 404:
                    raise NotFound(response, data)
                elif response.status_code >= 500:
                    raise BullhornServerError(response, data)
                else:
                    raise HTTPException(response, data)
            except OSError as e:
                # Connection reset by peer, so retry with exponential back-off
                if tries < 4 and e.errno in (54, 10054):
                    time.sleep(1 + tries * 2)
                    continue
                raise
        if response is not None:
            # Out of retries, so raise an appropriate exception
            if response.status_code >= 500:
                raise BullhornServerError(response, data)
            raise HTTPException(response, data)
        # Capture unhandled logic
        raise RuntimeError(response, "Unreachable code in HTTP handling")

    def ping(
        self,
    ) -> ping.Ping:
        """Ping used to test whether the client’s session is valid."""
        request = self.request(
            Route(
                "GET",
                self.rest_url + "ping",
            )
        )
        return request

    def get_candidates(
        self,
        query: str,
        fields: str,
    ) -> List[candidate.Candidate]:
        request = self.request(
            Route(
                "GET",
                self.rest_url + "search/Candidate?query={query}&fields={fields}",
                path_params={
                    "query": query,
                    "fields": fields,
                },
            )
        )
        return request

    def get_client_contacts(
        self,
        query: str,
        fields: str,
    ) -> List[client_contact.ClientContact]:
        request = self.request(
            Route(
                "GET",
                self.rest_url + "search/ClientContact?query={query}&fields={fields}",
                path_params={
                    "query": query,
                    "fields": fields,
                },
            )
        )
        return request

    def get_client_corporations(
        self,
        query: str,
        fields: str,
    ) -> List[client_corporation.ClientCorporation]:
        request = self.request(
            Route(
                "GET",
                self.rest_url
                + "search/ClientCorporation?query={query}&fields={fields}",
                path_params={
                    "query": query,
                    "fields": fields,
                },
            )
        )
        return request

    def get_corporate_users(
        self,
        query: str,
        fields: str,
    ) -> List[corporate_user.CorporateUser]:
        request = self.request(
            Route(
                "GET",
                self.rest_url + "search/CorporateUser?query={query}&fields={fields}",
                path_params={
                    "query": query,
                    "fields": fields,
                },
            )
        )
        return request

    def get_job_orders(
        self,
        query: str,
        fields: str,
    ) -> List[job_order.JobOrder]:
        request = self.request(
            Route(
                "GET",
                self.rest_url + "search/JobOrder?query={query}&fields={fields}",
                path_params={
                    "query": query,
                    "fields": fields,
                },
            )
        )
        return request

    def get_job_submissions(
        self,
        query: str,
        fields: str,
    ) -> List[job_submission.JobSubmission]:
        request = self.request(
            Route(
                "GET",
                self.rest_url + "search/JobSubmission?query={query}&fields={fields}",
                path_params={
                    "query": query,
                    "fields": fields,
                },
            )
        )
        return request

    def get_placements(
        self,
        query: str,
        fields: str,
    ) -> List[placement.Placement]:
        request = self.request(
            Route(
                "GET",
                self.rest_url + "search/Placement?query={query}&fields={fields}",
                path_params={
                    "query": query,
                    "fields": fields,
                },
            )
        )
        return request

    def get_placement_commissions(
        self,
        query: str,
        fields: str,
    ) -> List[placement_commission.PlacementCommission]:
        request = self.request(
            Route(
                "GET",
                self.rest_url
                + "search/PlacementCommission?query={query}&fields={fields}",
                path_params={
                    "query": query,
                    "fields": fields,
                },
            )
        )
        return request



from typing import List, Optional

from bullhorn.types import (
    agreement_line,
    billing_profile,
    branch,
    category,
    client_contact,
    client_corporation,
    corporate_user,
    job_code,
    location,
    opportunity,
    shift,
    workers_compensation_rate,
)


class JobOrder:
    id: int
    address: dict
    appointments: List[int]
    approvedPlacements: List[int]
    assignedUsers: List[int]
    benefits: str
    billRateCategoryID: int
    billingProfile: Optional[billing_profile.BillingProfile]
    bonusPackage: str
    branch: Optional[branch.Branch]
    branchCode: str
    businessSectors: List[int]
    categories: List[int]
    certificationGroups: List[int]
    certificationList: str
    certifications: List[int]
    clientBillRate: float
    clientContact: client_contact.ClientContact
    clientCorporation: client_corporation.ClientCorporation
    clientCorporationLine: Optional[agreement_line.AgreementLine]
    correlatedCustomDate1: Optional[int]
    correlatedCustomDate2: Optional[int]
    correlatedCustomDate3: Optional[int]
    correlatedCustomFloat1: Optional[float]
    correlatedCustomFloat2: Optional[float]
    correlatedCustomFloat3: Optional[float]
    correlatedCustomText1: Optional[str]
    correlatedCustomText2: Optional[str]
    correlatedCustomText3: Optional[str]
    correlatedCustomText4: Optional[str]
    correlatedCustomText5: Optional[str]
    correlatedCustomText6: Optional[str]
    correlatedCustomText7: Optional[str]
    correlatedCustomText8: Optional[str]
    correlatedCustomText9: Optional[str]
    correlatedCustomText10: Optional[str]
    costCenter: str
    customDate1: Optional[int]
    customDate2: Optional[int]
    customDate3: Optional[int]
    customFloat1: Optional[float]
    customFloat2: Optional[float]
    customFloat3: Optional[float]
    customFloat4: Optional[float]
    customFloat5: Optional[float]
    customFloat6: Optional[float]
    customFloat7: Optional[float]
    customFloat8: Optional[float]
    customInt1: Optional[int]
    customInt2: Optional[int]
    customInt3: Optional[int]
    customInt4: Optional[int]
    customInt5: Optional[int]
    customInt6: Optional[int]
    customInt7: Optional[int]
    customInt8: Optional[int]
    customText1: Optional[str]
    customText2: Optional[str]
    customText3: Optional[str]
    customText4: Optional[str]
    customText5: Optional[str]
    customText6: Optional[str]
    customText7: Optional[str]
    customText8: Optional[str]
    customText9: Optional[str]
    customText10: Optional[str]
    customTextBlock1: Optional[str]
    customTextBlock2: Optional[str]
    customTextBlock3: Optional[str]
    customTextBlock4: Optional[str]
    customTextBlock5: Optional[str]
    dateAdded: int
    dateClosed: Optional[int]
    dateEnd: Optional[int]
    dateLastExported: Optional[int]
    dateLastModified: int
    dateLastPublished: Optional[int]
    degreeList: str
    description: str
    durationWeeks: float
    educationDegree: str
    employmentType: str
    externalCategoryID: int
    externalID: str
    feeArrangement: float
    fileAttachments: List[int]
    hoursOfOperation: str
    hoursPerWeek: float
    interviews: List[int]
    isClientEditable: bool
    isDeleted: bool
    isInterviewRequired: bool
    isJobcastPublished: bool
    isOpen: bool
    isPublic: int
    jobBoardList: str
    jobCode: Optional[job_code.JobCode]
    jobOrderIntegrations: List[int]
    location: Optional[location.Location]
    markUpPercentage: float
    notes: List[int]
    numOpenings: int
    onSite: str
    opportunity: Optional[opportunity.Opportunity]
    optionsPackage: str
    owner: corporate_user.CorporateUser
    payRate: float
    placements: List[int]
    publicDescription: str
    publishedCategory: Optional[category.Category]
    publishedZip: str
    reasonClosed: str
    reportTo: str
    reportToClientContact: Optional[client_contact.ClientContact]
    responseUser: Optional[corporate_user.CorporateUser]
    salary: float
    salaryUnit: str
    sendouts: List[int]
    shift: Optional[shift.Shift]
    shifts: List[int]
    skillList: str
    skills: List[int]
    source: str
    specialties: List[int]
    startDate: int
    status: str
    submissions: List[int]
    tasks: List[int]
    taxRate: float
    taxStatus: str
    tearsheets: List[int]
    timeUnits: List[int]
    title: str
    travelRequirements: str
    type: int
    usersAssigned: str
    webResponses: List[int]
    willRelocate: bool
    willRelocateInt: int
    willSponsor: bool
    workersCompRate: Optional[workers_compensation_rate.WorkersCompensationRate]
    yearsRequired: int



from typing import Any, List, Optional

from bullhorn.types.person import Person


class ClientContact(Person):
    id: int
    address: Optional[str]
    businessSectors: List[int]
    category: int
    categories: List[int]
    certifications: str
    clientCorporation: int
    comments: str
    customDate1: Optional[int]
    customDate2: Optional[int]
    customDate3: Optional[int]
    customFloat1: Optional[float]
    customFloat2: Optional[float]
    customFloat3: Optional[float]
    customInt1: Optional[int]
    customInt2: Optional[int]
    customInt3: Optional[int]
    customObject1s: List[Any]
    customObject2s: List[Any]
    customObject3s: List[Any]
    customObject4s: List[Any]
    customObject5s: List[Any]
    customObject6s: List[Any]
    customObject7s: List[Any]
    customObject8s: List[Any]
    customObject9s: List[Any]
    customObject10s: List[Any]
    customText1: Optional[str]
    customText2: Optional[str]
    customText3: Optional[str]
    customText4: Optional[str]
    customText5: Optional[str]
    customText6: Optional[str]
    customText7: Optional[str]
    customText8: Optional[str]
    customText9: Optional[str]
    customText10: Optional[str]
    customText11: Optional[str]
    customText12: Optional[str]
    customText13: Optional[str]
    customText14: Optional[str]
    customText15: Optional[str]
    customText16: Optional[str]
    customText17: Optional[str]
    customText18: Optional[str]
    customText19: Optional[str]
    customText20: Optional[str]
    customTextBlock1: Optional[str]
    customTextBlock2: Optional[str]
    customTextBlock3: Optional[str]
    customTextBlock4: Optional[str]
    customTextBlock5: Optional[str]
    dateAdded: int
    dateLastModified: int
    dateLastVisit: Optional[int]
    description: str
    desiredCategories: str
    desiredSkills: str
    desiredSpecialties: str
    division: str
    email: str
    email2: Optional[str]
    email3: Optional[str]
    externalID: str
    fax: Optional[str]
    fax2: Optional[str]
    fax3: Optional[str]
    firstName: str
    isAnonymized: Optional[bool]
    isDayLightSavings: Optional[bool]
    isDeleted: bool
    lastName: str
    leads: List[int]
    linkedPerson: Person  # Person is a superclass of Candidate and ClientContact
    massMailOptOut: bool
    middleName: str
    mobile: Optional[str]
    name: str
    namePrefix: Optional[str]
    nameSuffix: Optional[str]
    nickName: Optional[str]
    numEmployees: int
    occupation: str
    office: str
    owner: int
    pager: Optional[str]
    password: str
    phone: str
    phone2: Optional[str]
    phone3: Optional[str]
    preferredContact: str
    referredByPerson: int
    reportToPerson: int
    secondaryAddress: Optional[str]
    secondaryOwners: List[int]
    skills: List[int]
    smsOptIn: bool
    source: str
    status: str
    timeZoneOffsetEST: int
    type: Optional[str]
    username: str
    salary: Optional[float]
    salaryUnit: str
    placements: List[int]
    publicDescription: str
    publishedCategory: int
    publishedZip: str
    reasonClosed: str
    reportTo: str
    reportToClientContact: int
    responseUser: int
    shifts: List[int]
    skillList: str
    skills: List[int]
    source: str
    specialties: List[int]
    startDate: int
    status: str
    submissions: List[int]
    tasks: List[int]
    taxRate: float
    taxStatus: str
    tearsheets: List[int]
    timeUnits: List[int]
    title: str
    travelRequirements: str
    type: int
    usersAssigned: str
    webResponses: List[int]
    willRelocate: bool
    willRelocateInt: int
    willSponsor: bool
    workersCompRate: int
    yearsRequired: int



from typing import Optional

from bullhorn.types import client_corporation, corporate_user


class Shift:
    id: int
    clientCorporation: Optional[client_corporation.ClientCorporation]
    dateLastModified: int
    dayOfWeek: Optional[int]
    endTime: Optional[int]
    htmlColorCode: Optional[str]
    isDefault: Optional[bool]
    isDeleted: bool
    lastModifyingUser: Optional[corporate_user.CorporateUser]
    migrateGUID: Optional[str]
    name: Optional[str]
    shortName: Optional[str]
    startTime: Optional[int]
    type: Optional[str]



from typing import List

from bullhorn import __version__
from bullhorn.client import BullhornClient
from bullhorn.types import (
    candidate,
    client_contact,
    client_corporation,
    corporate_user,
    job_order,
    job_submission,
    ping,
    placement,
    placement_commission,
)


def test_client_user_agent():
    # Initialise client
    bc = BullhornClient(
        token="12345_1234567_a12345bc-123a-45bc-67de-12345678910a",
        rest_url="https://rest123.bullhornstaffing.com/rest-services/1234/",
    )
    # Get user agent
    user_agent = bc.user_agent
    assert "Python wrapper for Bullhorn API" in user_agent


def test_client_ping(mocker):
    # Mock response
    return_value: ping.Ping = {"sessionExpires": 1234567891011}
    mocker.patch(
        "bullhorn.client.BullhornClient.request",
        return_value=return_value,
    )
    # Initialise client
    bc = BullhornClient(
        token="12345_1234567_a12345bc-123a-45bc-67de-12345678910a",
        rest_url="https://rest123.bullhornstaffing.com/rest-services/1234/",
    )
    # Get result
    result = bc.ping()
    assert "sessionExpires" in result


def test_client_get_candidates(mocker):
    # Mock response
    return_value: List[candidate.Candidate] = [
        {
            "id": 1234567891011,
            "firstName": "Example",
            "middleName": "C.",
            "lastName": "Name",
        },
    ]
    mocker.patch(
        "bullhorn.client.BullhornClient.request",
        return_value=return_value,
    )
    # Initialise client
    bc = BullhornClient(
        token="12345_1234567_a12345bc-123a-45bc-67de-12345678910a",
        rest_url="https://rest123.bullhornstaffing.com/rest-services/1234/",
    )
    # Get result
    result = bc.get_candidates(
        query="dateLastModified:{2023/01/01 TO *}",
        fields="id,firstName,middleName,lastName",
    )
    assert "firstName" in result[0]


def test_client_get_client_contacts(mocker):
    # Mock response
    return_value: List[client_contact.ClientContact] = [
        {
            "id": 1234567891011,
            "firstName": "Example",
            "middleName": "C.",
            "lastName": "Name",
        },
    ]
    mocker.patch(
        "bullhorn.client.BullhornClient.request",
        return_value=return_value,
    )
    # Initialise client
    bc = BullhornClient(
        token="12345_1234567_a12345bc-123a-45bc-67de-12345678910a",
        rest_url="https://rest123.bullhornstaffing.com/rest-services/1234/",
    )
    # Get result
    result = bc.get_client_contacts(
        query="dateLastModified:{2023/01/01 TO *}",
        fields="id,firstName,middleName,lastName",
    )
    assert "lastName" in result[0]


def test_client_get_client_corporations(mocker):
    # Mock response
    return_value: List[client_corporation.ClientCorporation] = [
        {
            "id": 1234567891011,
            "name": "Example Inc.",
        },
    ]
    mocker.patch(
        "bullhorn.client.BullhornClient.request",
        return_value=return_value,
    )
    # Initialise client
    bc = BullhornClient(
        token="12345_1234567_a12345bc-123a-45bc-67de-12345678910a",
        rest_url="https://rest123.bullhornstaffing.com/rest-services/1234/",
    )
    # Get result
    result = bc.get_client_corporations(
        query="dateLastModified:{2023/01/01 TO *}",
        fields="id,name",
    )
    assert "name" in result[0]


def test_client_get_corporate_users(mocker):
    # Mock response
    return_value: List[corporate_user.CorporateUser] = [
        {
            "id": 1234567891011,
            "firstName": "Example",
            "middleName": "C.",
            "lastName": "Name",
            "username": "ExampleCName",
        },
    ]
    mocker.patch(
        "bullhorn.client.BullhornClient.request",
        return_value=return_value,
    )
    # Initialise client
    bc = BullhornClient(
        token="12345_1234567_a12345bc-123a-45bc-67de-12345678910a",
        rest_url="https://rest123.bullhornstaffing.com/rest-services/1234/",
    )
    # Get result
    result = bc.get_corporate_users(
        query="dateLastModified:{2023/01/01 TO *}",
        fields="id,firstName,middleName,lastName,username",
    )
    assert "username" in result[0]


def test_client_get_job_orders(mocker):
    # Mock response
    return_value: List[job_order.JobOrder] = [
        {
            "id": 1234567891011,
            "description": "An example open job to be filled.",
            "owner": {
                "id": 1234567891011,
                "firstName": "Example",
                "middleName": "C.",
                "lastName": "Name",
                "username": "ExampleCName",
            },
        },
    ]
    mocker.patch(
        "bullhorn.client.BullhornClient.request",
        return_value=return_value,
    )
    # Initialise client
    bc = BullhornClient(
        token="12345_1234567_a12345bc-123a-45bc-67de-12345678910a",
        rest_url="https://rest123.bullhornstaffing.com/rest-services/1234/",
    )
    # Get result
    result = bc.get_job_orders(
        query="dateLastModified:{2023/01/01 TO *}",
        fields="id,description,owner",
    )
    assert "owner" in result[0]


def test_client_get_job_submissions(mocker):
    # Mock response
    return_value: List[job_submission.JobSubmission] = [
        {
            "id": 1234567891011,
            "endDate": 1234567891011,
            "owners": [
                {
                    "id": 1234567891011,
                    "firstName": "Example",
                    "middleName": "C.",
                    "lastName": "Name",
                    "username": "ExampleCName",
                },
            ],
            "startDate": 1234567891012,
        },
    ]
    mocker.patch(
        "bullhorn.client.BullhornClient.request",
        return_value=return_value,
    )
    # Initialise client
    bc = BullhornClient(
        token="12345_1234567_a12345bc-123a-45bc-67de-12345678910a",
        rest_url="https://rest123.bullhornstaffing.com/rest-services/1234/",
    )
    # Get result
    result = bc.get_job_submissions(
        query="dateLastModified:{2023/01/01 TO *}",
        fields="id,endDate,owners,startDate",
    )
    assert "owners" in result[0]


def test_client_get_placements(mocker):
    # Mock response
    return_value: List[placement.Placement] = [
        {
            "id": 1234567891011,
            "fee": 0.5,
        },
    ]
    mocker.patch(
        "bullhorn.client.BullhornClient.request",
        return_value=return_value,
    )
    # Initialise client
    bc = BullhornClient(
        token="12345_1234567_a12345bc-123a-45bc-67de-12345678910a",
        rest_url="https://rest123.bullhornstaffing.com/rest-services/1234/",
    )
    # Get result
    result = bc.get_placements(
        query="dateLastModified:{2023/01/01 TO *}",
        fields="id,fee",
    )
    assert "fee" in result[0]


def test_client_get_placement_commissions(mocker):
    # Mock response
    return_value: List[placement_commission.PlacementCommission] = [
        {
            "id": 1234567891011,
            "user": {
                "id": 1234567891011,
                "firstName": "Example",
                "middleName": "C.",
                "lastName": "Name",
                "username": "ExampleCName",
            },
            "commissionPercentage": 0.05,
        },
    ]
    mocker.patch(
        "bullhorn.client.BullhornClient.request",
        return_value=return_value,
    )
    # Initialise client
    bc = BullhornClient(
        token="12345_1234567_a12345bc-123a-45bc-67de-12345678910a",
        rest_url="https://rest123.bullhornstaffing.com/rest-services/1234/",
    )
    # Get result
    result = bc.get_placement_commissions(
        query="dateLastModified:{2023/01/01 TO *}",
        fields="id,user,commissionPercentage",
    )
    assert "commissionPercentage" in result[0]



from bullhorn import __version__, __title__, __author__, __license__
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    version=__version__,
    name=__title__,
    author=__author__,
    license=__license__,
    author_email="lewis@recruit-hub.com",
    description="Python wrapper for the Bullhorn REST API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/recruithub/bullhorn",
    packages=find_packages(),
    install_requires=[
        "requests==2.31.0",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
)



from typing import Any, Dict, List, Optional, Tuple, Union


class BullhornException(Exception):
    """Base exception class for the Bullhorn REST API wrapper library.

    This class serves as the base exception that can be caught to handle any exceptions raised from this library.
    """

    pass


class HTTPException(BullhornException):
    """Exception raised for HTTP-related errors when interacting with the Bullhorn REST API.

    This exception class contains information about the HTTP response, status code, and error message.
    """

    def __init__(
        self,
        response,
        message: Optional[Union[str, Dict[str, Any]]] = None,
    ):
        """
        Initialize an HTTPException object.

        Args:
            response (ClientResponse):
                The aiohttp ClientResponse object representing the HTTP response.

            message (Optional[Union[str, Dict[str, Any]]]):
                The error message associated with the exception.
                It can be either a string or a dictionary containing additional error details.

        Attributes:
            response (ClientResponse):
                The aiohttp ClientResponse object representing the HTTP response.

            status (int):
                The status code of the HTTP response.

            code (int):
                The error code associated with the exception.
                If the message is a dictionary, it is retrieved from the "code" key.
                Otherwise, it defaults to 0.

            text (str):
                The error message text associated with the exception.
                If the message is a dictionary, it is retrieved from the "message" key.
                Otherwise, it defaults to an empty string.

        Raises:
            None

        """
        self.response = response
        self.status: int = response.status_code
        self.code: int
        self.text: str

        if isinstance(message, dict):
            self.code = message.get("code", 0)
            base = message.get("message", "")
            errors = message.get("errors")
            self._errors: Optional[Dict[str, Any]] = errors

            if errors:
                errors = _flatten_error_dict(errors)
                helpful = "\n".join("In %s: %s" % t for t in errors.items())
                self.text = base + "\n" + helpful
            else:
                self.text = base
        else:
            self.text = message or ""
            self.code = 0

        fmt = f"{self.response.status_code} {self.response.reason} (error code: {self.code})"
        if len(self.text):
            fmt += f": {self.text}"

        super().__init__(fmt)


class Forbidden(HTTPException):
    """Exception that is raised when a 403 Forbidden status code occurs."""

    pass


class LoginFailure(HTTPException):
    """Exception that is raised when improper credentials are passed to log in."""

    pass


class NotFound(HTTPException):
    """Exception that is raised when a 404 Not Found status code occurs."""

    pass


class BullhornServerError(HTTPException):
    """Exception that is raised when a status code in the 500 range occurs."""

    pass


def _flatten_error_dict(d: Dict[str, Any], key: str = "") -> Dict[str, str]:
    """
    Flatten a nested error dictionary into a flat dictionary.

    This function takes a nested dictionary of errors and flattens it into a flat dictionary
    where each key represents the path to the error and the value is the corresponding error message.

    Args:
        d (Dict[str, Any]):
            The nested dictionary of errors.

        key (str, optional):
            The current key being processed in the recursive call.
            Defaults to an empty string.

    Returns:
        Dict[str, str]:
            A flat dictionary where each key represents the path to the error and the value is the error message.

    """
    # Iterate over nested dictionary and flatten into a flat dictionary
    items: List[Tuple[str, str]] = []
    for k, v in d.items():
        # Keep track of path to each error
        new_key = key + "." + k if key else k
        if isinstance(v, dict):
            try:
                # Extract error message from nested dictionary
                _errors: List[Dict[str, Any]] = v["_errors"]
            except KeyError:
                # Not at top level, so recursively call
                items.extend(_flatten_error_dict(v, new_key).items())
            else:
                items.append((new_key, " ".join(x.get("message", "") for x in _errors)))
        else:
            items.append((new_key, v))
    # Return items list as the flattened error dictionary
    return dict(items)



from typing import Optional


class CorporateUser:
    id: int
    companyName: str
    corporation: Optional[int]
    dateLastComment: Optional[int]
    dateLastModified: int
    email: Optional[str]
    email2: Optional[str]
    email3: Optional[str]
    enabled: bool
    externalEmail: str
    isDeleted: Optional[bool]
    isHidden: Optional[bool]
    isLockedOut: Optional[bool]
    lastName: Optional[str]
    masterUserID: int
    middleName: Optional[str]
    mobile: Optional[str]
    name: Optional[str]
    namePrefix: Optional[str]
    nameSuffix: Optional[str]
    nickName: Optional[str]
    occupation: Optional[str]
    status: Optional[str]
    userDateAdded: int
    userType: int
    username: str
